﻿using RTedge;
using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgFINS.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022, Programmed by Kazumi.M
/// 
/// 
///   -OMRON PLC data input functions
///   
/// --------------------------------------------------------------------------
/// </summary>

namespace EgFINS
{
    /// <summary>
    /// サービス本体
    /// </summary>
    public partial class EdgeService
    {
        /// <summary>
        /// ターゲットInputPLCのリストアップ処理、TagRefInの解釈
        /// </summary>
        /// <returns></returns>
        int TargetPlcInputListup()
        {
            // 無効定義はTagRefリストから削除したいので逆順に処理してリスト数変化を許容しています
            for (var i = EGAPI.EGFW.tagRef_IN_List.Count-1 ; i>=0 ; i--)
            {
                var tagrefitem = EGAPI.EGFW.tagRef_IN_List[i];

                int result;
                object test = null;
                String dip = "";
                int dna = 0;
                int da1 = 0;
                int da2 = 0;
                String cmd = "";

                try
                {
                    // RT-edgeにタグが存在するかテスト
                    result = EGAPI.EgTagRead(tagrefitem.Name, ref test);
                    if (result != 0)
                    {
                        // 実在しないedgeタグが指定されているエラーのログを出力してスキップ継続
                        edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("TagRefs_IN Name='{0}' is not exist, Ignored", tagrefitem.Name));

                        // Tagrefから削除します
                        EGAPI.EGFW.tagRef_IN_List.RemoveAt(i);
                        continue;
                    }

                    // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                    Match m = Regex.Match(tagrefitem.Address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<DNA>\d+):(?<DA1>\d+):(?<DA2>\d+)/(?<CMD>.*)$");
                    dip = m.Groups["IP"].Value;
                    dna = int.Parse(m.Groups["DNA"].Value);
                    da1 = int.Parse(m.Groups["DA1"].Value);
                    da2 = int.Parse(m.Groups["DA2"].Value);
                    cmd = m.Groups["CMD"].Value;

                    // DNA,DA1,DA2設定値範囲の妥当性を確認
                    if ((dna < 0) || (dna > 127))
                        throw new Exception();
                    if ((da1 < 0) || (da1 > 255))
                        throw new Exception();
                    if ((da2 < 0) || (da2 > 16))
                        throw new Exception();

                    // 新しい接続先IP指定を検出
                    bool newconnection = (PLCs.ContainsKey(dip) == false);
                    if (newconnection == true)
                    {
                        // FINS接続の新設（デフォルト通信サイクル=1000ms）
                        var newplc = new EgFINS.PLC(this , 1000);
                        newplc.fins = new FinsConnecton();
                        newplc.Ip = dip;
                        newplc.DNA = (byte)dna;
                        newplc.DA1 = (byte)da1;
                        newplc.DA2 = (byte)da2;

                        // 対象タグリストの新設
                        newplc.tagRefIn = new List<TagPair>();
                        newplc.tagRefOut = new List<TagPair>();
                        newplc.tagAllOutput = new List<TagOutput>();
                        newplc.tagOrderOutput = new List<TagOutput>();

                        // PLCリストに(接続情報,対象タグリスト)追加
                        PLCs.Add(dip, newplc);

                        // 新しい接続先のログ情報を出力
                        edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Defined IP={1} PORT={2} DNA={3} DA1={4} DA2={5}", PLCs.Count-1, newplc.Ip, ServiceFINS_PortNo, newplc.DNA, newplc.DA1, newplc.DA2));
                    }
                }
                catch
                {
                    // TagRefInに指令されたタグから有効なFINSアドレス書式を得られなかったログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("Incompatible address format '{0}', Ignored.", tagrefitem.Address));
                    continue;
                }

                // EdgeタグとPLCデータの紐づけ
                try
                {
                    // タグペア情報の組み立て
                    TagPair tagitem = new TagPair();
                    tagitem.cmd = cmd;
                    tagitem.EdgeTag = tagrefitem;

                    // PLCごとの対象タグリストに追加
                    PLCs[dip].tagRefIn.Add(tagitem);
                }
                catch
                {
                    // タグリンク失敗のログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("PLC#{0} Input tag linkage failed {1} << {2}", PLCs.Count-1, tagrefitem.Name, cmd));
                    continue;
                }

                // タグリンク成功のログ情報を出力
                edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Input tag linkaged {1} << {2}", PLCs.Count-1, tagrefitem.Name, cmd));
            }
            return 0;
        }

        /// <summary>
        /// PLCリードデータの更新（タイマイベント）
        /// </summary>
        public int PlcInput(PLC targetplc)
        {
            int result;
            FinsConnecton conn = targetplc.fins;

            // 読込対象のタグ定義が存在しないなら正常終了
            var tagcount = targetplc.tagRefIn.Count;
            if (tagcount == 0)
                return 0;

            // PLCのタグ数だけ繰り返してMultiple読込コマンド用のリクエストパラメータを作る
            FinsConnecton.fins_multidata_tp[] buf = new FinsConnecton.fins_multidata_tp[tagcount];
            int i = 0;
            foreach (var item in targetplc.tagRefIn)
            {
                // Edgeのタグデータ型に合わせてFINSリクエストのデータ型を分岐します
                buf[i].address = item.cmd;
                switch (item.EdgeTag.Type)
                {
                    case (ushort)EGDEFINE.egTagDataType.Boolean:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_BIT;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.SByte:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_INT16;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Byte:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_UINT16;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Int16:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_INT16;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.UInt16:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_UINT16;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Int32:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_INT32;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.UInt32:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_UINT32;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Float:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_FLOAT;
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Double:
                        buf[i].type = FinsConnecton.FINS_DATA_TYPE_DOUBLE;
                        break;
                }
                i++;
            }

            // FINSコマンドの発行, 一括して取り出すタイプ
            result = conn.ReadMultiple(ref buf, tagcount);
            targetplc.LastError = (result==0)? targetplc.LastError:(UInt16)result;
            if (result == 0)
            {
                // 一括して取り出したデータをEdgeタグに格納する
                i = 0;
                foreach (var item in targetplc.tagRefIn)
                {
                    var byteValue = new byte[8];
                    object edgevalue = null;
                    switch (item.EdgeTag.Type)
                    {
                        case (ushort)EGDEFINE.egTagDataType.Boolean:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (bool)(byteValue[0]!=0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.SByte:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (SByte)(byteValue[0]);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Byte:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (Byte)(byteValue[0]);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Int16:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToInt16(byteValue,0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.UInt16:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToUInt16(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Int32:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToInt32(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.UInt32:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToUInt32(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Float:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToSingle(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Double:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToDouble(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                    }
                    i++;
                }
                return 0;
            }
            else
            {
                // 通信異常などによりコマンドに失敗した
                return -1;
            }
        }
    }
}
