﻿using RTedge;
using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgFINS.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022, Programmed by Kazumi.M
/// 
/// 
///   -OMRON PLC data output functions
///   
/// --------------------------------------------------------------------------
/// </summary>

namespace EgFINS
{
    /// <summary>
    /// サービス本体
    /// </summary>
    public partial class EdgeService
    {
        /// <summary>
        /// ターゲットOutputPLCのリストアップ処理、TagRefInの解釈
        /// </summary>
        /// <returns></returns>
        int TargetPlcOutputListup()
        {
            // 無効定義はTagRefリストから削除したいので逆順に処理してリスト数変化を許容しています
            for (var i = EGAPI.EGFW.tagRef_OUT_List.Count - 1; i >= 0; i--)
            {
                var tagrefitem = EGAPI.EGFW.tagRef_OUT_List[i];

                int result;
                object test=null;
                String dip = "";
                int dna = 0;
                int da1 = 0;
                int da2 = 0;
                String cmd = "";

                try
                {
                    // RT-edgeにタグが存在するかテストする
                    result = EGAPI.EgTagRead(tagrefitem.Name, ref test);
                    if (result != 0)
                    {
                        // 実在しないedgeタグが指定されているエラーのログを出力
                        edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("TagRefs_OUT Name='{0}' is not exist, Ignored", tagrefitem.Name));

                        // Tagrefから削除します
                        EGAPI.EGFW.tagRef_OUT_List.RemoveAt(i);
                        continue;
                    }

                    // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                    Match m = Regex.Match(tagrefitem.Address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<DNA>\d+):(?<DA1>\d+):(?<DA2>\d+)/(?<CMD>.*)$");
                    dip = m.Groups["IP"].Value;
                    dna = int.Parse(m.Groups["DNA"].Value);
                    da1 = int.Parse(m.Groups["DA1"].Value);
                    da2 = int.Parse(m.Groups["DA2"].Value);
                    cmd = m.Groups["CMD"].Value;

                    // DNA,DA1,DA2設定値範囲の妥当性を確認
                    if ((dna < 0) || (dna > 127))
                        throw new Exception();
                    if ((da1 < 0) || (da1 > 255))
                        throw new Exception();
                    if ((da2 < 0) || (da2 > 16))
                        throw new Exception();

                    // 新しい接続先IP指定を検出
                    bool newconnection = (PLCs.ContainsKey(dip) == false);
                    if (newconnection == true)
                    {
                        // FINS接続の新設（デフォルト通信サイクル=1000）
                        var newplc = new EgFINS.PLC(this, 1000);
                        newplc.fins = new FinsConnecton();
                        newplc.Ip = dip;
                        newplc.DNA = (byte)dna;
                        newplc.DA1 = (byte)da1;
                        newplc.DA2 = (byte)da2;

                        // 対象タグリストの新設
                        newplc.tagRefIn = new List<TagPair>();
                        newplc.tagRefOut = new List<TagPair>();
                        newplc.tagAllOutput = new List<TagOutput>();
                        newplc.tagOrderOutput = new List<TagOutput>();

                        // PLCリストに(接続情報,対象タグリスト)追加
                        PLCs.Add(dip, newplc);

                        // 新しい接続先のログ情報を出力
                        edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Defined IP={1} PORT={2} DNA={3} DA1={4} DA2={5}", PLCs.Count-1, newplc.Ip, ServiceFINS_PortNo, newplc.DNA, newplc.DA1, newplc.DA2));
                    }
                }
                catch
                {
                    // TagRefOutに指令されたタグから有効なFINSアドレス書式を得られなかったログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("Incompatible address format '{0}', Ignored.", tagrefitem.Address));
                    continue;
                }

                // EdgeタグとPLCデータの紐づけ
                try
                {
                    // タグペア情報の組み立て
                    TagOutput tagitem = new TagOutput();
                    tagitem.cmd = cmd;
                    tagitem.EdgeTag = tagrefitem;

                    // PLCごとの対象タグリスト(tagrefOutリスト)に追加
                    PLCs[dip].tagRefOut.Add(tagitem);

                    // PLCごとの対象タグリスト(allputputリスト)に追加
                    PLCs[dip].tagAllOutput.Add(tagitem);
                }
                catch
                {
                    // タグリンク失敗のログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("PLC#{0} Output tag linkage failed {1} >> {2}", PLCs.Count-1, tagrefitem.Name, cmd));
                    continue;
                }

                // タグリンク成功のログ情報を出力
                edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Output tag linkaged {1} >> {2}", PLCs.Count-1, tagrefitem.Name, cmd));
            }
            return 0;
        }

        /// <summary>
        /// PLCライトデータの更新（タイマイベント）
        /// </summary>
        public int PlcOutput(PLC targetplc)
        {
            int result=0;
            FinsConnecton conn = targetplc.fins;

            // 書き込み対象のタグ定義が存在しないなら正常終了
            var tagcount = targetplc.tagAllOutput.Count;
            if (tagcount == 0)
                return 0;

            // 出力予約リストのクリア
            targetplc.tagOrderOutput.Clear();

            // PLCのタグ数だけ繰り返して
            foreach(var item in targetplc.tagAllOutput)
            {
                var address = item.cmd;
                var edgetagname = item.EdgeTag.Name;

                // Edgeタグの読み込み、失敗なら次へ
                object updatevalue = 0;
                result = EGAPI.EgTagRead(edgetagname, ref updatevalue);
                if (result != 0)
                    continue;

                if( item.value == null)
                {
                    // nullのときは、強制出力対象として出力予約リストへ追加
                    item.value = updatevalue;
                    targetplc.tagOrderOutput.Add(item);
                }
                else
                {
                    // 変化がないときは出力せずスキップ
                    Int64 p1 = Convert.ToInt64(updatevalue);
                    Int64 p2 = Convert.ToInt64(item.value);
                    if (p1 == p2)
                        continue;

                    // 出力対象として出力予約リストへ追加
                    item.value = updatevalue;
                    targetplc.tagOrderOutput.Add(item);
                }
            }

            // 出力予約リストにたまっているだけ回す
            foreach (var item in targetplc.tagOrderOutput)
            {
                // Edgeのタグデータ型に合わせてFINSコマンドリクエストが分岐します
                switch (item.EdgeTag.Type)
                {
                    case (ushort)EGDEFINE.egTagDataType.Boolean:
                        var blvalue = new Byte[1];
                        blvalue[0] = Convert.ToByte(item.value);
                        result = conn.WriteBit(item.cmd, ref blvalue, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.SByte:
                        var sbyvalue = new short[1];
                        sbyvalue[0] = Convert.ToInt16(item.value);
                        result = conn.WriteInt16(item.cmd, ref sbyvalue, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Byte:
                        var byvalue = new ushort[1];
                        byvalue[0] = Convert.ToUInt16(item.value);
                        result = conn.WriteUInt16(item.cmd, ref byvalue, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Int16:
                        var int16value = new short[1];
                        int16value[0] = Convert.ToInt16(item.value);
                        result = conn.WriteInt16(item.cmd, ref int16value, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.UInt16:
                        var uint16value = new ushort[1];
                        uint16value[0] = Convert.ToUInt16(item.value);
                        result = conn.WriteUInt16(item.cmd, ref uint16value, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Int32:
                        var int32value = new Int32[1];
                        int32value[0] = Convert.ToInt32(item.value);
                        result = conn.WriteInt32(item.cmd, ref int32value, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.UInt32:
                        var uint32value = new UInt32[1];
                        uint32value[0] = Convert.ToUInt32(item.value);
                        result = conn.WriteUInt32(item.cmd, ref uint32value, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Float:
                        var floatvalue = new UInt32[1];
                        // floatをbyte列に変換後、WORD列に変換
                        var byteFloats = BitConverter.GetBytes((float)item.value);
                        Buffer.BlockCopy(byteFloats, 0, floatvalue, 0, 4);
                        result = conn.WriteUInt32(item.cmd, ref floatvalue, 1);
                        break;
                    case (ushort)EGDEFINE.egTagDataType.Double:
                        var doublevalue = new UInt32[2];
                        // doubleをbyte列に変換後、WORD列に変換
                        var byteDoubles = BitConverter.GetBytes((double)item.value);
                        Buffer.BlockCopy(byteDoubles, 4, doublevalue, 4, 4);
                        Buffer.BlockCopy(byteDoubles, 0, doublevalue, 0, 4);
                        conn.WriteUInt32(item.cmd, ref doublevalue, 2);
                        break;
                }
                targetplc.LastError = (result == 0) ? targetplc.LastError : (UInt16)result;
            }
            return 0;
        }
    }
}
