﻿using RTedge;
using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Diagnostics;

/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgFINS.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022, Programmed by Kazumi.M
/// 
/// 
///   -RT-edge service main frame module
///   -Tag pair data class
///   -PLC management class
///   
/// --------------------------------------------------------------------------
/// </summary>

namespace EgFINS
{
    /// <summary>
    /// タグ関連付けのアイテム
    /// </summary>
    public class TagPair
    {
        public String cmd;              // デバイスアドレス
        public EgHgTagRef EdgeTag;      // Edgeタグ
    }

    /// <summary>
    /// アウトプット用のタグ関連付けアイテム
    /// </summary>
    public class TagOutput : TagPair
    {
        public object value;      // 出力したい値

        public TagOutput()
        {
            value = null;
        }
    }

    /// <summary>
    /// 接続先PLCのアイテム
    /// </summary>
    public class PLC
    {
        public Byte _Status;                 // PLC接続のステータス
        public Byte Status
        {
            set
            {
                // 変化があるならPLC毎のインジケータを更新する
                if (value != _Status)
                {
                    string tagname;
                    tagname = String.Format("SERVICE.{0}.PLC#{1}.Status", container.GetServiceName(), GetPlcNo());
                    _Status = value;
                    container.EGAPI.EgTagWrite(tagname, _Status);
                }
            }
            get { return _Status; }
        }
        public UInt16 _LastError;            // libfinsの最終エラーコード
        public UInt16 LastError
        {
            set
            {
                if (value != _LastError)
                {
                    string tagname;
                    tagname = String.Format("SERVICE.{0}.PLC#{1}.LastError", container.GetServiceName(), GetPlcNo());
                    _LastError = value;
                    container.EGAPI.EgTagWrite(tagname, _LastError);
                }
            }
            get { return _LastError; }
        }
        public String Ip;
        public Byte DNA;
        public Byte DA1;
        public Byte DA2;
        public FinsConnecton fins;          // FINS接続
        public List<TagPair> tagRefIn;      // Input対象タグ(tagrefIn)
        public List<TagPair> tagRefOut;     // Output対象タグ(tagrefOut)
        public List<TagOutput> tagAllOutput;   // Output対象タグ#2 前回出力値も保持している全ての出力
        public List<TagOutput> tagOrderOutput; // Output対象タグ#3 出力の予約リスト
        public bool Connected;              // 接続中は true
        public bool Connecting;             // 接続試行中はtrue
        EdgeService container;
        System.Timers.Timer FinsCycleTimer;         // PLCリフレッシュタイマ
        UInt32 CycleTime = 100;                     // リフレッシュ周期(ms)

        /// <summary>
        /// Constractor
        /// </summary>
        /// <param name="mom">EdgeServiceクラスインスタンスを与えてください</param>
        /// <param name="_CycleTime">リフレッシュサイクルタイム(ms)の指定</param>
        public PLC(EdgeService mom ,UInt32 _CycleTime)
        {
            container = mom;
            CycleTime = _CycleTime;
        }

        /// <summary>
        /// （全）PLCへの接続
        /// </summary>
        /// <returns></returns>
        public int Connect()
        {
            IndicatorCreate();

            // FINS接続要求
            Connecting = true;
            Connected = false;
            IndicatorUpdate();
            var err = fins.TcpConnect(Ip, container.ServiceFINS_PortNo
                                                , container.ServiceFINS_SNA, container.ServiceFINS_SA1, container.ServiceFINS_SA2
                                                , DNA, DA1, DA2
                                            );
            Connecting = false;
            Connected = (err == 0);
            IndicatorUpdate();

            // PLC管理番号をディクショナリ内から探す（確実に存在します）
            //plcNo = GetPlcNo(Ip);

            // 再接続時は全出力を強制
            ForceOutput();

            if (err == 0)
            {
                // PLC接続成功のログ情報を出力
//                edge_FW._PrintSysLog(container.GetServiceName(), "OPE", String.Format("PLC#{0} {1}/{2}:{3}:{4} Connection accepted.", plcNo, Ip, DNA, DA1, DA2));
                return 0;
            }
            else
            {
                // PLC接続失敗のログ情報を出力
                edge_FW._PrintSysLog(container.GetServiceName(), "ERR", String.Format("PLC#{0} {1}/{2}:{3}:{4} Connection refused.", GetPlcNo(), Ip, DNA, DA1, DA2));
                return -1;
            }
        }

        /// <summary>
        /// 表示上のPLC管理番号（XML定義出現順）を返す。PLCsリストとは逆順になっているので注意
        /// </summary>
        /// <returns>PLC管理番号</returns>
        public int GetPlcNo()
        {
            // ipに該当するPLC管理番号をディクショナリ内から探す
            int no = 0;
            foreach (var plc in container.PLCs)
            {
                if (plc.Value.Ip == Ip)
                    break;
                no++;
            }
            // 表示番号はPLCsリストと逆順になる。０始まり
            return container.PLCs.Count - no - 1;
        }

        /// <summary>
        /// （全）PLCの切断
        /// </summary>
        /// <returns></returns>
        public int Disconnect()
        {
            // PLC切断のログ情報を出力
            edge_FW._PrintSysLog(container.GetServiceName(), "OPE", String.Format("PLC#{0} {1}/{2}:{3}:{4} Disconnected.", GetPlcNo(), Ip, DNA, DA1, DA2));
            Connecting = false;
            Connected = false;
            IndicatorUpdate();

            // FINS通信の切断
            fins.Disconnect();
            return 0;
        }

        /// <summary>
        /// FINS通信状態インジケータの生成
        /// </summary>
        /// <returns></returns>
        int IndicatorCreate()
        {
            int result;
            string tagname;

            // タグを作ります
            tagname = String.Format("SERVICE.{0}.PLC#{1}.Status" ,container.GetServiceName() , GetPlcNo());
            result = container.EGAPI.EgTagCreate(tagname, (ushort)EGDEFINE.egTagDataType.Byte, 1, "", String.Format("PLC#{0} connection status", GetPlcNo()));

            tagname = String.Format("SERVICE.{0}.PLC#{1}.LastError", container.GetServiceName(), GetPlcNo());
            result = container.EGAPI.EgTagCreate(tagname, (ushort)EGDEFINE.egTagDataType.UInt16, 2, "", String.Format("PLC#{0} communication lasterror code", GetPlcNo()));

            return result;
        }

        /// <summary>
        /// FINS通信状態インジケータの更新
        /// </summary>
        /// <returns></returns>
        int IndicatorUpdate()
        {
            // PLC毎のステータスを更新
            if (Connecting)
                Status = 1;   // 接続試行中
            else if (Connected)
                Status = 2;   // 接続済通信中
            else
                Status = 0;   // 切断状態

            return 0;
        }

        /// <summary>
        /// リフレッシュタイマの初期化
        /// </summary>
        /// <returns></returns>
        public int RefleshInit(uint _CycleTime)
        {
            // タイマイベントが生成されていない時に限り生成
            if (FinsCycleTimer == null)
            {
                // リフレッシュサイクル(ms)指定を読み込む
                CycleTime = _CycleTime;

                //UInt32 CycleTime = container.ServiceFINS_PLC_Cycle[plcNo];

                // タイマイベントを生成する
                FinsCycleTimer = new System.Timers.Timer(CycleTime);
                FinsCycleTimer.Elapsed += RefleshEvent;
            }
            return 0;
        }

        /// <summary>
        /// リフレッシュタイマのスタート
        /// </summary>
        /// <returns></returns>
        public int RefleshStart()
        {
            // タイマが停止中だったら稼働させます
            if( !FinsCycleTimer.Enabled )
                FinsCycleTimer.Start();
            return 0;
        }

        /// <summary>
        /// リフレッシュタイマのストップ
        /// </summary>
        /// <returns></returns>
        public int RefleshStop()
        {
            // タイマが稼働中だったら止めます
            if (FinsCycleTimer.Enabled)
                FinsCycleTimer.Stop();
            return 0;
        }

        /// <summary>
        /// PLC入出力リフレッシュの根幹（タイマイベント）
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        private void RefleshEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            int result;

            // タイマー止める
            ((System.Timers.Timer)source).Stop();

            // 切断されているなら再接続
            if (!Connected)
            {
                result = Connect();
                if (result != 0)
                    goto end;
            }

            // データ更新有効じゃないなら無処理
            if (!container.ServiceIndRun)
                goto end;

            // PLCデータ入力
            result = container.PlcInput(this);

            if (result != 0)
            {
                // コマンドに失敗したので切断する
                Connected = false;
                Connecting = false;
                goto end;
            }

            // PLCデータ出力
            result = container.PlcOutput(this);

            if (result != 0)
            {
                // コマンドに失敗したので切断する(libfins内でdisconnect済なのでフラグだけです)
                Connected = false;
                Connecting = false;
                goto end;
            }

        end:
            // 次のタイマーイベントを許可
            FinsCycleTimer.Start();
        }

        /// <summary>
        /// 次回のリフレッシュサイクルで全出力を強制する
        /// </summary>
        /// <returns></returns>
        public int ForceOutput()
        {
            // 全ての出力要求データをnullに設定することで書き込みリクエストを用意させる
            foreach (var item in tagAllOutput)
            {
                item.value = null;
            }
            return 0;
        }
    }

    /// <summary>
    /// サービス本体
    /// </summary>
    public partial class EdgeService
    {
        public edge_API EGAPI;                                              // edge APIのインスタンス
        public Dictionary<string,PLC> PLCs = new Dictionary<string, PLC>(); // 接続PLCの管理リスト
        private System.Timers.Timer ServiceIndLiveTimer;                    // サービスインジケータ用のタイマ

        // FINS property
        public UInt16 ServiceFINS_PortNo = 9600;                            // FINS通信のデフォルトポート番号
        public Byte ServiceFINS_SNA = 0;
        public Byte ServiceFINS_SA1 = 0;
        public Byte ServiceFINS_SA2 = 0;

        // SPEC
        public const Byte MAX_PLC = 8;                                      // 対応できる最大PLC数
        public const UInt32 PlcCycleMsDefault = 1000;                       // PLCるフレッシュサイクル(ms)のデフォルト
        public List<UInt32> ServiceFINS_PLC_Cycle　= new List<uint>();      // PLC毎のサイクル周期(ms), XMLから読み取った結果が格納される

        // xmlで指定したサービスコンテナ名 EgInit後設定
        public string ServiceRealName = "";
        // サービスコンテナ名デフォルト値 EgInit前で使用
        public string DefServiceName = "EgFINS";
        /// <summary>
        /// インジケータタグ".Run" / Service Run indicator
        /// </summary>
        bool _ServiceIndRun;
        public bool ServiceIndRun
        {
            set
            {
                _ServiceIndRun = value;
                String TagName = String.Format("SERVICE.{0}.Run", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndRun; }
        }

        /// <summary>
        /// インジケータタグ".Status" / Service Status indicator
        /// </summary>
        byte _ServiceIndStatus;
        byte ServiceIndStatus
        {
            set
            {
                _ServiceIndStatus = value;
                String TagName = String.Format("SERVICE.{0}.Status", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndStatus; }
        }

        /// <summary>
        /// インジケータタグ".Error" / Service Error indicator
        /// </summary>
        bool _ServiceIndError;
        bool ServiceIndError
        {
            set
            {
                _ServiceIndError = value;
                String TagName = String.Format("SERVICE.{0}.Error", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndError; }
        }

        /// <summary>
        /// インジケータタグ".Live" / ervice Live counter indicator
        /// </summary>
        UInt32 _ServiceIndLive;
        UInt32 ServiceIndLive
        {
            set
            {
                _ServiceIndLive = value;
                String TagName = String.Format("SERVICE.{0}.Live", GetServiceName());
                EGAPI.EgTagWrite(TagName, value);
            }
            get { return _ServiceIndLive; }
        }


        /// <summary>
        /// 自己のサービス名を返す
        /// </summary>
        /// <returns></returns>
        public String GetServiceName()
        {
            if (ServiceRealName.Length > 0) //XMLで指定したサービス名 EgInit後に設定 設定がない場合はデフォルト値を返す
                return ServiceRealName;
            return DefServiceName; //"EgFINS"
        }

        /// <summary>
        /// サービスプロパティのXML解釈処理
        /// </summary>
        /// <returns></returns>
        public int GetProperty()
        {
            object Value = 0;
            int ret;
            String TagName;

            //
            // プロパティタグ取得".Port" / ServiceFINS_PortNo / Type5 = UInt16 / FINS port numnber
            //
            TagName = String.Format("SERVICE.{0}.Port", GetServiceName());
            ret = EGAPI.EgTagRead(TagName, ref Value);
            if (ret == 0)
                ServiceFINS_PortNo = Convert.ToUInt16(Value);
            else
            {
                // 定義が存在しなければデフォルト値でタグを用意するよ
                ret = EGAPI.EgTagCreate(TagName, (ushort)EGDEFINE.egTagDataType.UInt16, 2, "", "FINS port number");
                ret = EGAPI.EgTagWrite(TagName, ServiceFINS_PortNo);
            }

            //
            // プロパティタグ取得".SNA" / ServiceFINS_SNA / Type3 = Byte / FINS Source net
            //
            TagName = String.Format("SERVICE.{0}.SNA", GetServiceName());
            ret = EGAPI.EgTagRead(TagName, ref Value);
            if (ret == 0)
                ServiceFINS_SNA = Convert.ToByte(Value);
            else
            {
                // 定義が存在しなければデフォルト値でタグを用意するよ
                ret = EGAPI.EgTagCreate(TagName, (ushort)EGDEFINE.egTagDataType.Byte, 1, "", "FINS Source net");
                ret = EGAPI.EgTagWrite(TagName, ServiceFINS_SNA);
            }

            //
            // プロパティタグ取得".SA1" / ServiceFINS_SA1 / Type3 = Byte / FINS Source Node
            //
            TagName = String.Format("SERVICE.{0}.SA1", GetServiceName());
            ret = EGAPI.EgTagRead(TagName, ref Value);
            if (ret == 0)
                ServiceFINS_SA1 = Convert.ToByte(Value);
            else
            {
                // 定義が存在しなければデフォルト値でタグを用意するよ
                ret = EGAPI.EgTagCreate(TagName, (ushort)EGDEFINE.egTagDataType.Byte, 1, "", "FINS Source Node");
                ret = EGAPI.EgTagWrite(TagName, ServiceFINS_SA1);
            }

            //
            // プロパティタグ取得".SA2" / ServiceFINS_SA2 / Type3 = Byte / FINS Source Unit
            //
            TagName = String.Format("SERVICE.{0}.SA2", GetServiceName());
            ret = EGAPI.EgTagRead(TagName, ref Value);
            if (ret == 0)
                ServiceFINS_SA2 = Convert.ToByte(Value);
            else
            {
                // 定義が存在しなければデフォルト値でタグを用意するよ
                ret = EGAPI.EgTagCreate(TagName, (ushort)EGDEFINE.egTagDataType.Byte, 1, "", "FINS Source Unit");
                ret = EGAPI.EgTagWrite(TagName, ServiceFINS_SA2);
            }

            return 0;
        }

        /// <summary>
        /// PLCリフレッシュサイクル(ms)のプロパティを取得する処理
        /// 定義がないときはデフォルト値でEdgeタグを作ります
        /// </summary>
        /// <param name="numplc">何個のPLCが定義されているかを指定する</param>
        private void GetPlcCycleProperty(byte numplc)
        {
            int i;
            int ret;
            object Value = 0;
            string TagName;
            UInt32 plccyclems;

            for (i = 0; i < numplc; i++)
            {
                // プロパティタグの存在を読み取ってみることで確認
                TagName = String.Format("SERVICE.{0}.PLC#{1}.Cycle", GetServiceName(), i);      // Uint32型
                ret = EGAPI.EgTagRead(TagName, ref Value);
                if (ret == 0)
                {
                    // 指定が存在しても許容範囲でなければデフォルト値を採用する
                    plccyclems = Convert.ToUInt32(Value);
                    if( (plccyclems<100)||(plccyclems>65535))
                    {
                        plccyclems = PlcCycleMsDefault;
                        ret = EGAPI.EgTagWrite(TagName, plccyclems);
                    }

                    // リフレッシュサイクルとして採用
                    ServiceFINS_PLC_Cycle.Add(plccyclems);
                }
                else
                {
                    // 見つからない時はデフォルト値でタグを作り採用する
                    plccyclems = PlcCycleMsDefault;
                    ServiceFINS_PLC_Cycle.Add(plccyclems);

                    // プロパティタグを自動生成します
                    ret = EGAPI.EgTagCreate(TagName, (ushort)EGDEFINE.egTagDataType.UInt32, 4, "", String.Format("PLC#{0} reflesh Cycletime in ms" ,i));
                    ret = EGAPI.EgTagWrite(TagName, plccyclems);
                }
            }
        }

        /// <summary>
        /// インジケータタグの生成
        /// </summary>
        /// <returns></returns>
        public int CreateIndicater()
        {
            int ret = 0;

            // インジケータタグを生成します
            ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Status", GetServiceName()) ,(ushort)EGDEFINE.egTagDataType.Byte, 1, "", "service status");
            ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Error", GetServiceName()), (ushort)EGDEFINE.egTagDataType.Boolean, 1, "", "service last error");
            ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Run", GetServiceName()), (ushort)EGDEFINE.egTagDataType.Boolean, 1, "", "service runnning");
            ret |= EGAPI.EgTagCreate(String.Format("SERVICE.{0}.Live", GetServiceName()), (ushort)EGDEFINE.egTagDataType.UInt32, 4, "", "service live counter");

            return ret;
        }

        /// <summary>
        /// サービス起動時処理
        /// </summary>
        /// <returns></returns>
        public int Startup()
        {
            // RT-edgeの準備
            EGAPI = new edge_API();

            // RT-edgeユーザハンドラ登録
            edge_API.EDGECONFIG config = EGAPI.EDGE_CONFIG_DEFAULT;
            config.mInstMessFunc = MsgHandler;

            // RT-edgeフレームワーク初期化
            var ret = EGAPI.EgInit( config );
            if (ret != 0)
                edge_FW._PrintSysLog( GetServiceName(), "ERR", "Failed to initialize. Please check the environment.");

            // サービス名の設定
            ServiceRealName = EGAPI.EGFW.ServiceRealName;

            // インジケータタグの準備
            CreateIndicater();

            // XMLプロパティタグ設定の取り出し
            GetProperty();

            // FINSマスタ設定のログ情報を出力
            edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("Source IP={0} PORT={1} SNA={2} SA1={3} SA2={4}", "0.0.0.0" ,ServiceFINS_PortNo , ServiceFINS_SNA, ServiceFINS_SA1, ServiceFINS_SA2));

            // 対象PLCのリスティング,TagRefIn/TagRefOutの解釈
            TargetPlcInputListup();
            TargetPlcOutputListup();

            // PLCリフレッシュサイクルのプロパティタグ設定の取り出し、初期設定
            var numplc = Convert.ToByte(PLCs.Count);
            GetPlcCycleProperty(numplc);

            // Winsockの初期化
            WinSock.Startup();

            // （全）PLCへの接続
            foreach (var plc in PLCs)
            {
                plc.Value.Connect();

                // サイクル周期設定(ms)の取り出し
                var plcno = plc.Value.GetPlcNo();
                var cycle = ServiceFINS_PLC_Cycle[plcno];

                // データリフレッシュの初期化
                plc.Value.RefleshInit(cycle);

                // データリフレッシュ開始
                plc.Value.RefleshStart();
            }

           // ライブインジケータタグの周期更新スタート / Service Live counter timer
            ServiceIndLiveTimer = new System.Timers.Timer(1000);
            ServiceIndLiveTimer.Elapsed += OnTimedLiveIndicaterEvent;
            ServiceIndLiveTimer.Start();

            return 0;
        }

        /// <summary>
        /// Liveインジケータの更新（タイマイベント）
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        private void OnTimedLiveIndicaterEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            try
            {
                // インジケータ更新
                ((System.Timers.Timer)source).Stop();
                ServiceIndLive++;

                // コンテナのステータスインジケータを更新、通信対象PLCのいずれかに障害があれば .Error=true にします
                int errors = 0;
                foreach( var plc in PLCs)
                {
                    if (plc.Value.Status != 2)
                        errors++;
                }
                ServiceIndError = (errors > 0);
            }
            finally
            {
                // 次のタイマーイベントを許可
                ((System.Timers.Timer)source).Start();
            }
        }

        /// <summary>
        /// サービス終了時の処理
        /// </summary>
        /// <returns></returns>
        public int Cleanup()
        {
            // FINS通信(PLC接続)の全面カット
            foreach(var plc in PLCs)
            {
                plc.Value.RefleshStop();
                plc.Value.Disconnect();
            }

            // Liveカウンタタイマのストップ
            ServiceIndLiveTimer.Stop();

            // Winsockのクリンナップ
            WinSock.Cleanup();

            // RT-edgeフレームワーク終了
            EGAPI.EgFinalize();

            // プロセス終了
//            Application.Exit();
            var myself = Process.GetCurrentProcess();
            myself.CloseMainWindow();
            return 0;
        }

        /// <summary>
        /// サービスメッセージハンドラ
        /// </summary>
        /// <param name="senderName"></param>
        /// <param name="messNo"></param>
        /// <param name="paramArray"></param>
        /// <returns></returns>
        public int MsgHandler(string senderName, int messNo, byte[] paramArray)
        {
            try
            {
                // 届いたEdgeメッセージに合わせて処理分岐しますよ
                switch (messNo)
                {
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_RUN:            // サービス開始指令
                        ServiceIndRun = true;
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_STOP:           // サービス終了指令
                        ServiceIndRun = false;
                        Cleanup();
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_PAUSE:          // サービス一時停止指令
                        ServiceIndRun = false;
                        break;
                    case (int)EGDEFINE.EGMSG.EM_SERVICE_UPDATE:			// サービスアップデート指令
                        break;
                    default:
                        return -1;
                }
                return 0;
            }
            catch (Exception ex)
            {
                edge_FW._PrintSysLog( GetServiceName(), "ERR", "FATAL ERROR : {0}", System.Reflection.MethodBase.GetCurrentMethod().Name + " : " + ex.Message);
                return -1;
            }
        }
    }
}
