﻿using RTedge;
using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgMC.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022, Programmed by K.M
/// 
/// 
///   -OMRON PLC data input functions
///   
/// --------------------------------------------------------------------------
/// </summary>

namespace EgMC
{
    /// <summary>
    /// サービス本体
    /// </summary>
    public partial class EdgeService
    {
        /// <summary>
        /// ターゲットInputPLCのリストアップ処理、TagRefInの解釈
        /// </summary>
        /// <returns></returns>
        int TargetPlcInputListup()
        {
            // 無効定義はTagRefリストから削除したいので逆順に処理してリスト数変化を許容しています
            for (var i = EGAPI.EGFW.tagRef_IN_List.Count-1 ; i>=0 ; i--)
            {
                var tagrefitem = EGAPI.EGFW.tagRef_IN_List[i];

                int result;
                object test = null;
                String dip = "";
                int Device = 0;
                ushort usport = 0;
                int NetNo = 0;
                int PCNo = 0;
                int IONo = 0;
                int Station = 0;
                int Adress = 0;
                String cmd = "";
                string hexString = "";
                byte bDevice = 0;
                string stDevice = "";
                try
                {
                    // RT-edgeにタグが存在するかテスト
                    result = EGAPI.EgTagRead(tagrefitem.Name, ref test);
                    if (result != 0)
                    {
                        // 実在しないedgeタグが指定されているエラーのログを出力してスキップ継続
                        edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("TagRefs_IN Name='{0}' is not exist, Ignored", tagrefitem.Name));

                        // Tagrefから削除します
                        EGAPI.EGFW.tagRef_IN_List.RemoveAt(i);
                        continue;
                    }

                    // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                    Match m = Regex.Match(tagrefitem.Address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");
                    dip = m.Groups["IP"].Value;
                    hexString = m.Groups["NetNo"].Value;
                    hexString = hexString.Replace("h","");
                    NetNo = Int32.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    hexString = m.Groups["PCNo"].Value;
                    hexString = hexString.Replace("h", "");
                    PCNo = Int32.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    hexString = m.Groups["IONo"].Value;
                    hexString = hexString.Replace("h", "");
                    IONo = int.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    hexString = m.Groups["Station"].Value;
                    hexString = hexString.Replace("h", "");
                    Station = int.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    Adress = int.Parse(m.Groups["Adress"].Value);
                    stDevice = m.Groups["Device"].Value;

                    if ((NetNo < 0) || (NetNo > 255))
                        throw new Exception();
                    if ((PCNo < 0) || (PCNo > 255))
                        throw new Exception();
                    if ((IONo < 0) || (IONo > 65535))
                        throw new Exception();
                    if ((Station < 0) || (Station > 255))
                        throw new Exception();
                    if ((Adress < 0) || (Adress > 16777215))
                        throw new Exception();

                    bDevice = ChkDevType(stDevice);
                    if (bDevice == 0)
                        throw new Exception();



                    // 新しい接続先IP指定を検出
                    bool newconnection = (PLCs.ContainsKey(dip) == false);
                    if (newconnection == true)
                    {
                        // MC接続の新設（デフォルト通信サイクル=1000ms）
                        var newplc = new EgMC.PLC(this , 1000);
                        newplc.MC = new MCConnecton();
                        newplc.Ip = dip;
                        newplc.port = usport;

                        // 対象タグリストの新設
                        newplc.tagRefIn = new List<TagPair>();
                        newplc.tagRefOut = new List<TagPair>();
                        newplc.tagAllOutput = new List<TagOutput>();
                        newplc.tagOrderOutput = new List<TagOutput>();

                        // PLCリストに(接続情報,対象タグリスト)追加
                        PLCs.Add(dip, newplc);

                        // 新しい接続先のログ情報を出力
                        edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Defined IP={1} Port={2} ", PLCs.Count-1, newplc.Ip, newplc.port));
                    }
                }
                catch
                {
                    // TagRefInに指令されたタグから有効なMCアドレス書式を得られなかったログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("Incompatible address format '{0}', Ignored.", tagrefitem.Address));
                    continue;
                }

                // EdgeタグとPLCデータの紐づけ
                try
                {
                    // タグペア情報の組み立て
                    TagPair tagitem = new TagPair();
                    tagitem.cmd = cmd;
                    tagitem.EdgeTag = tagrefitem;

                    // PLCごとの対象タグリストに追加
                    PLCs[dip].tagRefIn.Add(tagitem);
                }
                catch
                {
                    // タグリンク失敗のログ情報を出力
                    edge_FW._PrintSysLog(GetServiceName(), "WARN", String.Format("PLC#{0} Input tag linkage failed {1} << {2}", PLCs.Count-1, tagrefitem.Name, cmd));
                    continue;
                }

                // タグリンク成功のログ情報を出力
                edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("PLC#{0} Input tag linkaged {1} << {2}", PLCs.Count-1, tagrefitem.Name, cmd));
            }
            return 0;
        }




        /// <summary>
        /// PLCリードデータの更新（タイマイベント）
        /// </summary>
        public int PlcInput(PLC targetplc)
        {
#if DEBUG
            var sw = new System.Diagnostics.Stopwatch();
            // 計測開始
            sw.Start();
#else
#endif

            int result;
            MCConnecton conn = targetplc.MC;

            // 読込対象のタグ定義が存在しないなら正常終了
            var tagcount = targetplc.tagRefIn.Count;
            if (tagcount == 0)
                return 0;

            // PLCのタグ数だけ繰り返してMultiple読込コマンド用のリクエストパラメータを作る
            MCConnecton.MC_multidata_tp[] buf = new MCConnecton.MC_multidata_tp[tagcount];
            int i = 0;

            foreach (var item in targetplc.tagRefIn)
            {
                // Edgeのタグデータ型に合わせてMCリクエストのデータ型を分岐します
                //buf[i].address = item.cmd;
                buf[i].address = item.EdgeTag.Address;
                buf[i].type = item.EdgeTag.Type;
                //switch (item.EdgeTag.Type)
                //{
                //    case (ushort)EGDEFINE.egTagDataType.Boolean:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_BIT;
                //        iDlen = 1;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.SByte:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_INT16;
                //        iDlen = 2;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.Byte:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_UINT16;
                //        iDlen = 2;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.Int16:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_INT16;
                //        iDlen = 2;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.UInt16:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_UINT16;
                //        iDlen = 2;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.Int32:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_INT32;
                //        iDlen = 4;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.UInt32:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_UINT32;
                //        iDlen = 4;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.Float:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_FLOAT;
                //        iDlen = 4;
                //        break;
                //    case (ushort)EGDEFINE.egTagDataType.Double:
                //        buf[i].type = MCConnecton.MC_DATA_TYPE_DOUBLE;
                //        iDlen = 8;
                //        break;
                //}

                //result = conn.ReadMCAll(ref buf[i], iDlen, item.EdgeTag.Type, ServiceMC_Frame, ServiceMC_Timer);
                i++;
            }
            result = conn.ReadMCAll(ref buf, targetplc.frame, targetplc.timer, i);
            targetplc.LastError = (result == 0) ? targetplc.LastError : (UInt16)result;
            if (result == 0)
            {
                i = 0;
                foreach (var item in targetplc.tagRefIn)
                {

                    var byteValue = new byte[8];
                    object edgevalue = null;
                    switch (item.EdgeTag.Type)
                    {
                        case (ushort)EGDEFINE.egTagDataType.Boolean:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (bool)(byteValue[0] != 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.SByte:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (SByte)(byteValue[0]);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Byte:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = (Byte)(byteValue[0]);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Int16:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToInt16(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.UInt16:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToUInt16(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Int32:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToInt32(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.UInt32:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToUInt32(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Float:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToSingle(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                        case (ushort)EGDEFINE.egTagDataType.Double:
                            byteValue = BitConverter.GetBytes(buf[i].value);
                            edgevalue = BitConverter.ToDouble(byteValue, 0);
                            result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
                            break;
                    }


                    i++;
                }
            }
            else
            {
                // 通信異常などによりコマンドに失敗した
                return -1;
            }

            // MCコマンドの発行, 一括して取り出すタイプ
            //result = conn.ReadMultiple(ref buf, tagcount);
            //result = conn.ReadMC(ref buf, 0,0);
            //targetplc.LastError = (result==0)? targetplc.LastError:(UInt16)result;
            //if (result == 0)
            //{
            //    // 一括して取り出したデータをEdgeタグに格納する
            //    i = 0;
            //    foreach (var item in targetplc.tagRefIn)
            //    {
            //        var byteValue = new byte[8];
            //        object edgevalue = null;
            //        switch (item.EdgeTag.Type)
            //        {
            //            case (ushort)EGDEFINE.egTagDataType.Boolean:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = (bool)(byteValue[0]!=0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.SByte:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = (SByte)(byteValue[0]);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.Byte:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = (Byte)(byteValue[0]);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.Int16:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToInt16(byteValue,0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.UInt16:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToUInt16(byteValue, 0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.Int32:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToInt32(byteValue, 0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.UInt32:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToUInt32(byteValue, 0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.Float:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToSingle(byteValue, 0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //            case (ushort)EGDEFINE.egTagDataType.Double:
            //                byteValue = BitConverter.GetBytes(buf[i].value);
            //                edgevalue = BitConverter.ToDouble(byteValue, 0);
            //                result = EGAPI.EgTagWrite(item.EdgeTag.Name, edgevalue);
            //                break;
            //        }
            //        i++;
            //    }
            //    return 0;
            //}
            //else
            //{
            //    // 通信異常などによりコマンドに失敗した
            //    return -1;
            //}

#if DEBUG
            // 計測停止
            sw.Stop();
            TimeSpan elapsed = sw.Elapsed;
            float eMili = (float)elapsed.TotalMilliseconds;
            float eMicr = eMili * 1000;

            //TimeSpan ts = sw.Elapsed;
            // タグリンク成功のログ情報を出力
            //edge_FW._PrintSysLog(GetServiceName(), "OPE", String.Format("EgMC PlcInput processing time {0}時間,{1}分,{2}秒,{3}ミリ秒,{4}ミリ秒", ts.Hours, ts.Minutes, ts.Seconds, ts.Milliseconds, sw.ElapsedMilliseconds));
            result = EGAPI.EgTagWrite("SERVICE.EgMC.ReadCheck", eMicr);

#else
#endif
            return 0;
        }
    }
}
