﻿using System;
using System.Runtime.InteropServices; // DLL Import
using System.Net.Sockets;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Threading;
/// <summary>
/// --------------------------------------------------------------------------
///   RT-edge system / Container
/// 
///   EgMC.exe(W)
///       OMROM PLC series data connector service
/// 
///   Copyright(c) Micronet Corp 2022,
/// 
/// 
///   -Winsock Wrapper module
///   
///             Windows socket (winsock) initializer
///     
///   
///   -MC protocol libraly Wrapper module
///   
///            libMC --- Opensource software MIT Licensed
///                        Copyright(c) 2016 Lammert Bies
///            
///            Non-blocking customized by K.M
/// --------------------------------------------------------------------------
/// </summary>
namespace EgMC
{
    /// <summary>
    /// WinSockラッパー
    /// </summary>
    public class WinSock
    {
        [StructLayout(LayoutKind.Sequential, Pack = 4)]
        struct WSAData
        {
            public short wVersion;
            public short wHighVersion;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 257)]
            public string szDescription;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 129)]
            public string szSystemStatus;
            public UInt16 iMaxSockets;
            public UInt16 iMaxUdpDg;
            public int lpVendorInfo;
        }

        /// <summary>
        /// かの有名なWSAStartup
        /// </summary>
        /// <param name="wVersionRequested"></param>
        /// <param name="lpWSAData"></param>
        /// <returns></returns>
        [DllImport("wsock32.dll")]
        static extern int WSAStartup([In] short wVersionRequested, [Out] out WSAData lpWSAData);

        /// <summary>
        /// 終了時に使用
        /// </summary>
        /// <returns></returns>
        [DllImport("wsock32.dll")]
        static extern int WSACleanup();

        /// <summary>
        /// Winsockの使用開始で使います
        /// </summary>
        static public void Startup()
        {
            WSAData dummy = new WSAData();
            WSAStartup(0x0202, out dummy);
        }

        /// <summary>
        /// Winsockの使用を終えたときに使用
        /// </summary>
        static public void Cleanup()
        {
            WSACleanup();
        }
    }

    /// <summary>
    /// LibMCラッパー
    /// </summary>
    public class MCConnecton
    {

        /// <summary>
        /// PLC構成情報
        /// </summary>
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct MC_cpudata_tp
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 21)]
            public string model;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 21)]
            public string version;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 40)]
            public char[] system_block;
            public int largest_em_bank;
            public int program_area_size;
            public int iom_size;
            public int number_of_dm_words;
            public int timer_counter_size;
            public int em_non_file_memory_size;
            public int memory_card_size;
            public int num_sysmac_bus_masters;
            public int num_racks;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
            public UInt16[] bus_unit_id;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
            public bool[] bus_unit_present;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
            public bool[] dip_switch;
            public byte memory_card_type;
        };

        /// <summary>
        /// MultiRead用のパラメータ構造体
        /// </summary>
        [StructLayout(LayoutKind.Sequential , CharSet=CharSet.Ansi)]
        public struct MC_multidata_tp
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 12)]
            public string address;
            public int type;
            public UInt64 value;
        };

        /// <summary>
        /// MultiReadで使用するMC既定のデータ型
        /// </summary>
        public const int MC_DATA_TYPE_NONE = 0;			/* No data type						*/
        public const int MC_DATA_TYPE_INT16 = 1;			/* 16 bit signed integer				*/
        public const int MC_DATA_TYPE_INT32 = 2;			/* 32 bit signed integer				*/
        public const int MC_DATA_TYPE_UINT16 = 3;			/* 16 bit unsigned integer				*/
        public const int MC_DATA_TYPE_UINT32 = 4;			/* 32 bit unsigned integer				*/
        public const int MC_DATA_TYPE_FLOAT = 15;			/* 32 bit floating point value				*/
        public const int MC_DATA_TYPE_DOUBLE = 16;		/* 64 bit floating point value				*/
        public const int MC_DATA_TYPE_BIT = 17;			/* Single bit						*/
        public const int MC_DATA_TYPE_BIT_FORCED = 18;	/* Single bit with forced status			*/
        public const int MC_DATA_TYPE_WORD_FORCED = 19;	/* 16 bit word with for each bit the forced status	*/
        public const int MC_DATA_TYPE_LAST = 19;			/* Last defined data type				*/

        /* 接続先サーバー情報 */
        private string m_serverAddress;
        private int m_serverPort;
        private int m_readTimeout;
        private int m_writeTimeout;

        /* 状態フラグ */
        public bool m_connected;

        /* TCP通信ハンドル */
        private TcpClient m_client;
        private NetworkStream m_tcpStream;

        /* 定数値 */
        private  int HDR_SIZE = 11;
        private enum commType { READ, WRITE };



        /// <summary>
        /// サーバー情報初期化
        /// </summary>
        /// <param name="serverAddress">サーバーアドレス</param>
        /// <param name="serverPort">サーバーポート番号</param>
        /// <param name="readTimeout">受信タイムアウト[ms]</param>
        /// <param name="writeTimeout">送信タイムアウト[ms]</param>
        /// <returns></returns>
        public bool SetConnectInfo(string serverAddress, int serverPort, int readTimeout = 1000, int writeTimeout = 1000)
        {
            // サーバー情報を更新
            m_serverAddress = serverAddress;
            m_serverPort = serverPort;
            m_readTimeout = readTimeout;
            m_writeTimeout = writeTimeout;

            // 「切断」状態に初期化
            m_connected = false;

            // エラー処理を追加してfalseを返すのがベスト
            return true;
        }

        /// <summary>
        /// PLCへの接続 TCP
        /// </summary>
        /// <returns>正常終了時0</returns>
        public int TcpConnect()
        {
            try
            {
                // サーバーと接続
                // 接続完了するまでブロッキングする
                m_client = new System.Net.Sockets.TcpClient(m_serverAddress, m_serverPort);

                // 「接続」状態に更新
                m_connected = true;

                // ネットワークストリームを取得
                m_tcpStream = m_client.GetStream();

                // 送受信タイムアウト時間を設定
                m_tcpStream.ReadTimeout = m_readTimeout;
                m_tcpStream.WriteTimeout = m_writeTimeout;
            }
            catch (Exception ex)
            {
                // 接続失敗
                Console.WriteLine($"{System.DateTime.Now.ToString("[yyyy/MM/dd HH:mm:ss]")}【TCPClient】Connect() : ERROR !!! {ex.Message}");
                m_tcpStream?.Close();
                m_client?.Close();
                m_connected = false;

                return 1;
            }





            return 0;
        }

        /// <summary>
        /// PLCへの接続 UDP
        /// </summary>
        /// <param name="address">PLC側のIPアドレス</param>
        /// <param name="port">MC共通ポート番号</param>
        /// <param name="local_net">RTedge側のネットワークアドレス</param>
        /// <param name="local_node">RTedge側のノードアドレス</param>
        /// <param name="local_unit">RTedge側のユニットNo</param>
        /// <param name="remote_net">PLC側のネットワークアドレス</param>
        /// <param name="remote_node">PLC側のノードアドレス</param>
        /// <param name="remote_unit">PLC側のユニットNo</param>
        /// <returns>正常終了時0</returns>
        //public int UdpConnect(String address, UInt16 port, Byte local_net, Byte local_node, Byte local_unit, Byte remote_net, Byte remote_node, Byte remote_unit)
        //{
        //    int error_val;
        //    sys = MClib_udp_connect(IntPtr.Zero, address, port, local_net, local_node, local_unit, remote_net, remote_node, remote_unit, out error_val, 1);

        //    if (error_val == 0x0000)
        //    {
        //        MC_cpudata_tp cpudata = new MC_cpudata_tp();
        //        MClib_cpu_unit_data_read(sys, ref cpudata);
        //    }
        //    return error_val;
        //}

        /// <summary>
        /// PLCの切断
        /// </summary>
        public void Disconnect()
        {
            m_tcpStream?.Close();
            m_client?.Close();
            m_connected = false;
        }
        /// <summary>
        /// ランダム読出しコマンド
        /// </summary>
        /// <param name="item">リード先を示した構造体配列</param>
        /// <param name="usFrame">フレーム　4or3</param>
        /// <param name="usTimer">タイマー値</param>
        /// <param name="item_count">配列数</param>
        /// <returns>正常終了時0、異常時は内部で切断されます</returns>
        public int ReadMCAll(ref MC_multidata_tp[] item, ushort usFrame, ushort usTimer, int item_count)
        {
            
            int sendSize = 0;
            short reqSize = 0;
            int m_sirialNo = 0;
            int i = 0;
            int ii = 0;
            int iii = 0;
            byte[] sendData;

            short wordNum = 0;
            short dwordNum = 0;
            short doubleNum = 0;
            string stNetNo = "";
            string stIONo = "";
            string stPCNo = "";
            string stStation = "";
            string[] stDevice = new string[item_count*2]; 
            int[] iAdress = new int[item_count*2];

            //ワード情報取得
            for (i = 0; i < item_count; i++)
            {
                // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                Match m = Regex.Match(item[i].address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");
                if (i==0)
                {
                    stNetNo = m.Groups["NetNo"].Value;
                    stPCNo = m.Groups["PCNo"].Value;
                    stIONo = m.Groups["IONo"].Value;
                    stStation = m.Groups["Station"].Value;
                }
                // Edgeのタグデータ型に合わせてMCコマンドリクエストが分岐します
                switch (item[i].type)
                {
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Boolean:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.SByte:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Byte:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Int16:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.UInt16:
                        stDevice[ii] = m.Groups["Device"].Value;
                        iAdress[ii] = int.Parse(m.Groups["Adress"].Value);
                        wordNum++;
                        ii++;
                        break;
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Int32:
                        break;
                    case (ushort)RTedge.EGDEFINE.egTagDataType.UInt32:
                        break;
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Float:
                        break;
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Double:
                        break;
                }




            }

            //ダブルワード情報取得
            for (i = 0; i < item_count; i++)
            {
                // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                Match m = Regex.Match(item[i].address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");
                // Edgeのタグデータ型に合わせてMCコマンドリクエストが分岐します
                switch (item[i].type)
                {
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Int32:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.UInt32:
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Float:
                        stDevice[ii] = m.Groups["Device"].Value;
                        iAdress[ii] = int.Parse(m.Groups["Adress"].Value);
                        ii++;
                        dwordNum++;
                        break;
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Double:
                        break;
                }
            }

            //ダブル情報取得
            for (i = 0; i < item_count; i++)
            {
                // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
                Match m = Regex.Match(item[i].address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");
                // Edgeのタグデータ型に合わせてMCコマンドリクエストが分岐します
                switch (item[i].type)
                {
                    case (ushort)RTedge.EGDEFINE.egTagDataType.Double:
                        stDevice[ii] = m.Groups["Device"].Value;
                        iAdress[ii] = int.Parse(m.Groups["Adress"].Value);
                        ii++;


                        stDevice[ii] = m.Groups["Device"].Value;
                        iAdress[ii] = int.Parse(m.Groups["Adress"].Value);
                        iAdress[ii] = iAdress[ii] + 2;
                        ii++;
                        doubleNum++;

                        break;
                }
            }

            ii = 0;

            switch (usFrame)
            {
                case 3://3Eフレーム
                    HDR_SIZE = 11;
                    sendSize = HDR_SIZE + 6 + (wordNum + dwordNum + doubleNum*2) * 4;
                    reqSize = (short)(sendSize - 9);
                    sendData = new byte[sendSize];

                    // サブヘッダ部
                    sendData[ii++] = 0x50;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部

                    sendData[ii++] = byte.Parse(stNetNo, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    sendData[ii++] = byte.Parse(stPCNo, System.Globalization.NumberStyles.HexNumber);                                 // PC番号
                    sendData[ii++] = (byte)((short.Parse(stIONo, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);        // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(stIONo, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);        // 〃 (上位)
                    sendData[ii++] = byte.Parse(stStation, System.Globalization.NumberStyles.HexNumber);                              // 要求先ユニット局番号

                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)
                    //sendData[ii++] = 0x10;                              // CPU監視タイマ
                    //sendData[ii++] = 0x00;                              // 〃 (上位)


                    // データ部 QnA互換4Eフレーム (CPU内蔵Ethernetポート向け)
                    // ※ 0406(複数ブロック一括読出)は出来ないことに注意
                    sendData[ii++] = 0x03;    // コマンド
                    sendData[ii++] = 0x04;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)
                    sendData[ii++] = (byte)wordNum;    // ワードアクセス点数
                    sendData[ii++] = (byte)(dwordNum + doubleNum*2);    // ダブルワードアクセス点数

                    iii = 0;
                    for (i = 0; i < wordNum; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }
                    for (i = 0; i < dwordNum; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }

                    for (i = 0; i < doubleNum*2; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }

                    break;
                case 4://4Eフレーム
                    HDR_SIZE = 15;
                    sendSize = HDR_SIZE + 6 + (wordNum + dwordNum + doubleNum*2) * 4; 
                    reqSize = (short)(sendSize - 13);
                    sendData = new byte[sendSize];

                    //// サブヘッダ部
                    sendData[ii++] = 0x54;
                    sendData[ii++] = 0x00;
                    m_sirialNo++;
                    sendData[ii++] = (byte)((m_sirialNo & 0x00ff) >> 0);     // シリアルNO
                    sendData[ii++] = (byte)((m_sirialNo & 0xff00) >> 8);     // シリアルNO
                    sendData[ii++] = 0x00;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部
                    sendData[ii++] = byte.Parse(stNetNo, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    sendData[ii++] = byte.Parse(stPCNo, System.Globalization.NumberStyles.HexNumber);                                 // PC番号
                    sendData[ii++] = (byte)((short.Parse(stIONo, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);        // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(stIONo, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);        // 〃 (上位)
                    sendData[ii++] = byte.Parse(stStation, System.Globalization.NumberStyles.HexNumber);                              // 要求先ユニット局番号
                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = 0x03;    // コマンド
                    sendData[ii++] = 0x04;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)

                    sendData[ii++] = (byte)wordNum;    // ワードアクセス点数
                    sendData[ii++] = (byte)(dwordNum + doubleNum * 2);    // ダブルワードアクセス点数

                    iii = 0;
                    for (i = 0; i < wordNum; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }
                    for (i = 0; i < dwordNum; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }

                    for (i = 0; i < doubleNum * 2; i++)
                    {
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x0000ff) >> 0);    // 先頭デバイス
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0x00ff00) >> 8);    // 〃
                        sendData[ii++] = (byte)(((iAdress[iii]) & 0xff0000) >> 16);   // 〃 (上位)
                        sendData[ii++] = ChkDevType(stDevice[iii]);                             // デバイスコード
                        iii++;

                    }


                    break;
                default:
                    return -1;
            }





            // 送信電文(メモリ書込) 送信
            if (Send(sendData) != 0)
            {
                return -1;
            }

            // 読出周期だけスリープ
            //Thread.Sleep(usTimer);

            // 応答待ち
            int recvExpectSize = HDR_SIZE + (wordNum*2 + dwordNum*4 + doubleNum * 8);
            byte[] recvCommData = new byte[recvExpectSize];
            int recvSize = Receive(recvCommData);

            byte[] recvData8 = new byte[8];

            if (recvSize > 0)
            {
                ii = 0;
                for (i = 0; i < item_count; i++)
                {

                    switch (item[i].type)
                    {
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Boolean:
                            item[i].value = (ulong)(recvCommData[HDR_SIZE]);
                            ii = ii + 2;
                            break;
                        case (ushort)RTedge.EGDEFINE.egTagDataType.SByte:
                            if (recvCommData[HDR_SIZE + 1] == 255)
                            {
                                item[i].value = (ulong)(recvCommData[HDR_SIZE + ii] * -1);
                            }
                            else
                            {
                                item[i].value = (ulong)(recvCommData[HDR_SIZE + ii]);
                            }
                            ii = ii + 2;
                            break;
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Byte:
                            item[i].value = (ulong)(recvCommData[HDR_SIZE + ii]);
                            ii = ii + 2;
                            break;
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Int16:
                            item[i].value = (ulong)(recvCommData[HDR_SIZE + ii] + recvCommData[HDR_SIZE + ii + 1] * 256);
                            ii = ii + 2;
                            break;
                        case (ushort)RTedge.EGDEFINE.egTagDataType.UInt16:
                            item[i].value = (ulong)(recvCommData[HDR_SIZE + ii] + recvCommData[HDR_SIZE + ii + 1] * 256);
                            ii = ii + 2;
                            break;
                    }
                }

                for (i = 0; i < item_count; i++)
                {

                    switch (item[i].type)
                    {
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Int32:
                        case (ushort)RTedge.EGDEFINE.egTagDataType.UInt32:
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Float:
                            item[i].value = (ulong)(recvCommData[HDR_SIZE + ii] + recvCommData[HDR_SIZE + ii + 1] * 256 + recvCommData[HDR_SIZE + ii + 2] * 256 * 256 + recvCommData[HDR_SIZE + ii + 3] * 256 * 256 * 256);
                            ii = ii + 4;
                            break;
                    }
                }
                for (i = 0; i < item_count; i++)
                {

                    switch (item[i].type)
                    {
                        case (ushort)RTedge.EGDEFINE.egTagDataType.Double:
                            ulong ul = BitConverter.ToUInt64(recvCommData, HDR_SIZE + ii);
                            item[i].value = ul;
                            ii = ii + 8;
                            break;
                    }
                }


                return 0;
            }

            return -1;
        }

        /// <summary>
        /// 一括読出しコマンド
        /// </summary>
        /// <param name="item">リード先を示した構造体配列</param>
        /// <param name="wordNum">バイト数</param>
        /// <param name="type">値のタイプ</param>
        /// <param name="usFrame">フレーム　4or3</param>
        /// <param name="usTimer">タイマー値</param>
        /// <returns>正常終了時0、異常時は内部で切断されます</returns>
        public int ReadMC(ref MC_multidata_tp item, short wordNum, ushort type, ushort usFrame, ushort usTimer)
        {
            int sendSize = 0;
            short reqSize = 0;
            string hexString = "";
            int m_sirialNo = 0;
            int ii = 0;
            byte[] sendData;
            // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
            Match m = Regex.Match(item.address, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");

            switch (usFrame)
            {
                case 3://3Eフレーム
                    HDR_SIZE = 11;
                    sendSize = HDR_SIZE + 10;
                    reqSize = (short)(sendSize - 9);
                    sendData = new byte[sendSize];

                    // サブヘッダ部
                    sendData[ii++] = 0x50;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部

                    hexString = m.Groups["NetNo"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    hexString = m.Groups["PCNo"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // PC番号
                    hexString = m.Groups["IONo"].Value;
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);       // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);       // 〃 (上位)
                    hexString = m.Groups["Station"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // 要求先ユニット局番号
                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)
                    //sendData[ii++] = 0x10;                              // CPU監視タイマ
                    //sendData[ii++] = 0x00;                              // 〃 (上位)


                    // データ部 QnA互換4Eフレーム (CPU内蔵Ethernetポート向け)
                    // ※ 0406(複数ブロック一括読出)は出来ないことに注意
                    sendData[ii++] = 0x01;    // コマンド
                    sendData[ii++] = 0x04;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x0000ff) >> 0);    // 先頭デバイス
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x00ff00) >> 8);    // 〃
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0xff0000) >> 16);   // 〃 (上位)

                    sendData[ii++] = ChkDevType(m.Groups["Device"].Value);                             // デバイスコード
                    sendData[ii++] = (byte)((wordNum & 0x00ff) >> 0);                                  // デバイス点数
                    sendData[ii++] = (byte)((wordNum & 0xff00) >> 8);                                  // 〃 (上位)



                    break;
                case 4://4Eフレーム
                    HDR_SIZE = 15;
                    sendSize = HDR_SIZE + 10;
                    reqSize = (short)(sendSize - 13);
                    sendData = new byte[sendSize];

                    //// サブヘッダ部
                    sendData[ii++] = 0x54;
                    sendData[ii++] = 0x00;
                    m_sirialNo++;
                    sendData[ii++] = (byte)((m_sirialNo & 0x00ff) >> 0);     // シリアルNO
                    sendData[ii++] = (byte)((m_sirialNo & 0xff00) >> 8);     // シリアルNO
                    sendData[ii++] = 0x00;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部
                    hexString = m.Groups["NetNo"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    hexString = m.Groups["PCNo"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // PC番号
                    hexString = m.Groups["IONo"].Value;
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);       // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);       // 〃 (上位)
                    hexString = m.Groups["Station"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // 要求先ユニット局番号


                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = 0x01;    // コマンド
                    sendData[ii++] = 0x04;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x0000ff) >> 0);    // 先頭デバイス
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x00ff00) >> 8);    // 〃
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0xff0000) >> 16);   // 〃 (上位)
                    hexString = m.Groups["Device"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);// デバイスコード
                    sendData[ii++] = (byte)((wordNum & 0x00ff) >> 0);                                  // デバイス点数
                    sendData[ii++] = (byte)((wordNum & 0xff00) >> 8);                                  // 〃 (上位)

                    break;
                default:
                    return -1;
            }





            // 送信電文(メモリ書込) 送信
            if (Send(sendData) != 0)
            {
                return -1;
            }

            // 読出周期だけスリープ
            //Thread.Sleep(usTimer);

            // 応答待ち
            int recvExpectSize = HDR_SIZE + wordNum * 2;
            byte[] recvCommData = new byte[recvExpectSize];
            int recvSize = Receive(recvCommData);


            if (recvSize > 0)
            {
                switch (wordNum)
                {
                    case 1:
                        item.value = (ulong)(recvCommData[HDR_SIZE]);
                        break;
                    case 2:
                        if (type == 2)
                        {
                            if (recvCommData[HDR_SIZE + 1] == 255)
                            {
                                item.value = (ulong)(recvCommData[HDR_SIZE] * -1);
                            }
                            else
                            {
                                item.value = (ulong)(recvCommData[HDR_SIZE]);
                            }
                        }
                        if (type == 3)
                        {
                            item.value = (ulong)(recvCommData[HDR_SIZE]);
                        }
                        if (type == 4)
                        {
                            item.value = (ulong)(recvCommData[HDR_SIZE] + recvCommData[HDR_SIZE + 1] * 256);
                        }
                        if (type == 5)
                        {
                            item.value = (ulong)(recvCommData[HDR_SIZE] + recvCommData[HDR_SIZE + 1] * 256);
                        }

                        break;
                    case 4:
                        item.value = (ulong)(recvCommData[HDR_SIZE] + recvCommData[HDR_SIZE + 1] * 256 + recvCommData[HDR_SIZE + 2] * 256 * 256 + recvCommData[HDR_SIZE + 3] * 256 * 256 * 256);
                        break;
                    case 8:
                        ulong ul = BitConverter.ToUInt64(recvCommData, HDR_SIZE);
                        item.value = ul;
                        break;
                }
                return 0;
            }

            return -1;
        }

        /// <summary>
        /// 一括書き込みコマンド（効率よいコマンドを構成してくれる）
        /// </summary>
        /// <param name="item">書込み先</param>
        /// <param name="wordNum">書込みバイト数</param>
        /// <param name="data">書き込む値</param>
        /// <param name="usFrame">フレーム</param>
        /// <param name="usTimer">タイマー値</param>
        /// <returns>正常終了時0、異常時は内部で切断されます</returns>
        public int WriteMC(string item, short wordNum, byte[] data, ushort usFrame, ushort usTimer)
        {
            int sendSize = 0;
            short reqSize = 0;
            string hexString = "";
            int m_sirialNo = 0;
            int ii = 0;

            byte[] sendData;
            // アドレス書式 "999.999.999.999/9:9:9/CIO999" の解釈
            //Match m = Regex.Match(item, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])):(?<PORT>\d+)/(?<NetNo>[0-9A-Fa-f]+):(?<PCNo>[0-9A-Fa-f]+):(?<IONo>[0-9A-Fa-f]+):(?<Station>[0-9A-Fa-f]+)/(?<Adress>\d+):(?<Device>[0-9A-Fa-f]+)/(?<Form>.+?)E");
            Match m = Regex.Match(item, @"^(?<IP>(([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([1-9]?[0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))/(?<NetNo>[0-9A-Fa-fh]+):(?<PCNo>[0-9A-Fa-fh]+):(?<IONo>[0-9A-Fa-fh]+):(?<Station>[0-9A-Fa-fh]+)/(?<Adress>\d+):(?<Device>.+)");

            switch (usFrame)
            {
                case 3://3Eフレーム
                    HDR_SIZE = 11;
                    sendSize = HDR_SIZE + 10 + wordNum*2;
                    reqSize = (short)(sendSize - 9);
                    sendData = new byte[sendSize];

                    // サブヘッダ部
                    sendData[ii++] = 0x50;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部

                    hexString = m.Groups["NetNo"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    hexString = m.Groups["PCNo"].Value;                                                                                 // PC番号
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    hexString = m.Groups["IONo"].Value;
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);       // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);       // 〃 (上位)
                    hexString = m.Groups["Station"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // 要求先ユニット局番号
                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)


                    // データ部 QnA互換4Eフレーム (CPU内蔵Ethernetポート向け)
                    // ※ 0406(複数ブロック一括読出)は出来ないことに注意
                    sendData[ii++] = 0x01;    // コマンド
                    sendData[ii++] = 0x14;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x0000ff) >> 0);    // 先頭デバイス
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x00ff00) >> 8);    // 〃
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0xff0000) >> 16);   // 〃 (上位)
                    sendData[ii++] = ChkDevType(m.Groups["Device"].Value);                             // デバイスコード
                    sendData[ii++] = (byte)((wordNum & 0x00ff) >> 0);                                  // デバイス点数
                    sendData[ii++] = (byte)((wordNum & 0xff00) >> 8);                                  // 〃 (上位)



                    break;
                case 4://4Eフレーム
                    HDR_SIZE = 15;
                    sendSize = HDR_SIZE + 10 + wordNum * 2;
                    reqSize = (short)(sendSize - 13);
                    sendData = new byte[sendSize];

                    //// サブヘッダ部
                    sendData[ii++] = 0x54;
                    sendData[ii++] = 0x00;
                    m_sirialNo++;
                    sendData[ii++] = (byte)((m_sirialNo & 0x00ff) >> 0);     // シリアルNO
                    sendData[ii++] = (byte)((m_sirialNo & 0xff00) >> 8);     // シリアルNO
                    sendData[ii++] = 0x00;
                    sendData[ii++] = 0x00;

                    // Qヘッダ部

                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // ネットワーク№
                    hexString = m.Groups["PCNo"].Value;                                                                                 // PC番号
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);
                    hexString = m.Groups["IONo"].Value;
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0x00ff) >> 0);       // 要求先ユニット I/O番号
                    sendData[ii++] = (byte)((short.Parse(hexString, System.Globalization.NumberStyles.HexNumber) & 0xff00) >> 8);       // 〃 (上位)
                    hexString = m.Groups["Station"].Value;
                    sendData[ii++] = byte.Parse(hexString, System.Globalization.NumberStyles.HexNumber);                                // 要求先ユニット局番号
                    sendData[ii++] = (byte)((reqSize & 0x00ff) >> 0);     // 要求データ長
                    sendData[ii++] = (byte)((reqSize & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = (byte)((usTimer & 0x00ff) >> 0);     // CPU監視タイマ
                    sendData[ii++] = (byte)((usTimer & 0xff00) >> 8);     // 〃 (上位)
                    sendData[ii++] = 0x01;    // コマンド
                    sendData[ii++] = 0x14;    // 〃 (上位)
                    sendData[ii++] = 0x00;    // サブコマンド
                    sendData[ii++] = 0x00;    // 〃 (上位)
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x0000ff) >> 0);    // 先頭デバイス
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x00ff00) >> 8);    // 〃
                    sendData[ii++] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0xff0000) >> 16);   // 〃 (上位)
                    sendData[ii++] = ChkDevType(m.Groups["Device"].Value);                             // デバイスコード
                    sendData[ii++] = (byte)((wordNum & 0x00ff) >> 0);                                  // デバイス点数
                    sendData[ii++] = (byte)((wordNum & 0xff00) >> 8);                                  // 〃 (上位)

                    break;
                default:
                    return -1;
            }






            if (wordNum == 1)
            {
                sendData[ii++] = data[0];     // 書込データ

            }
            if (wordNum == 2)
            {
                sendData[ii++] = data[0];     // 書込データ
                sendData[ii++] = data[1];     // 書込データ

            }
            if (wordNum == 4)
            {
                sendData[ii++] = data[0];     // 書込データ
                sendData[ii++] = data[1];     // 書込データ
                sendData[ii++] = data[2];     // 書込データ
                sendData[ii++] = data[3];     // 書込データ

            }
            if (wordNum == 8)
            {
                sendData[ii++] = data[0];     // 書込データ
                sendData[ii++] = data[1];     // 書込データ
                sendData[ii++] = data[2];     // 書込データ
                sendData[ii++] = data[3];     // 書込データ
                sendData[ii++] = data[4];     // 書込データ
                sendData[ii++] = data[5];     // 書込データ
                sendData[ii++] = data[6];     // 書込データ
                sendData[ii++] = data[7];     // 書込データ

            }



            // 送信電文(メモリ書込) 送信
            if (Send(sendData) != 0)
            {
                return -1;
            }

            // 読出周期だけスリープ
            //Thread.Sleep(usTimer);

            // 応答待ち
            int recvExpectSize = HDR_SIZE;
            byte[] recvCommData = new byte[recvExpectSize + 59];
            int recvSize = Receive(recvCommData);
            if (recvSize > 0)
            { return 0; }

            return -1;
        }

        /// <summary>
        /// 通信電文送信
        /// </summary>
        /// <param name="data"></param>
        public int Send(byte[] data)
        {
            try
            {
                // データ送信開始
                m_tcpStream.Write(data, 0, data.Length);

                // 送信成功
            }
            catch (Exception ex)
            {
                // 送信失敗

                // 「切断」状態に更新
                m_connected = false;

                // クライアント初期化
                m_tcpStream?.Close();
                m_client?.Close();
                return -1;
            }
            return 0;
        }

        /// <summary>
        /// 通信電文受信
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public int Receive(byte[] data)
        {
            int receiveSize = 0;
            int receiveCnt = 0;
            try
            {
                while (receiveCnt < 1000)
                {
                    // データ受信開始
                    while (m_tcpStream.DataAvailable)
                    {
                        receiveSize += m_tcpStream.Read(data, 0, data.Length);
                    }

                    // 受信成功
                    if (receiveSize > 0)
                    {
                        if (receiveCnt > 0)
                        {
                            receiveCnt = receiveCnt + 1000;

                        }
                        receiveCnt = receiveCnt + 1000;
                    }
                    else
                    {
                        Thread.Sleep(1);
                        receiveCnt++;
                    }
                }

            }
            catch (System.IO.IOException ex)
            {
                // タイムアウト
                Console.WriteLine($"{System.DateTime.Now.ToString("[yyyy/MM/dd HH:mm:ss]")}【TCPClient】Receive() : 受信タイムアウト");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"{System.DateTime.Now.ToString("[yyyy/MM/dd HH:mm:ss]")}【TCPClient】Receive() : ERROR !!! {ex.Message}");

                // 「切断」状態に更新
                m_connected = false;

                // クライアント初期化
                m_tcpStream?.Close();
                m_client?.Close();

                throw ex;
            }

            return receiveSize;
        }


        /// <summary>
        /// Multipleリードコマンド（効率よいコマンドを構成してくれる）
        /// </summary>
        /// <param name="item">リード先を示した構造体配列</param>
        /// <param name="num_item">配列数</param>
        /// <returns>正常終了時0、異常時は内部で切断されます</returns>
        public int ReadMultiple(ref MC_multidata_tp[] item, int num_item)
        {
            var size = Marshal.SizeOf(typeof(MC_multidata_tp)) * num_item;
            var ptr = Marshal.AllocCoTaskMem(size);
            IntPtr dst = ptr;
            int i;
            int ret = 0;
            for (i = 0; i < num_item; i++)
            {
                Marshal.StructureToPtr(item[i], dst, false);
                dst = (IntPtr)((long)dst + Marshal.SizeOf(item[i]));
            }
            //ret = MClib_multiple_memory_area_read(sys, ptr, num_item);
            if (ret == 0)
            {
                IntPtr src = ptr;
                for (i = 0; i < num_item; i++)
                {
                    item[i] = Marshal.PtrToStructure<MC_multidata_tp>(src);
                    src = (IntPtr)((long)src + Marshal.SizeOf(item[i]));
                }
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int WriteBit(String start, ref Byte[] data, int num_bit, ushort usFrame, ushort usTimer)
        {
            //byte[] sendData = new byte[2];
            //sendData[0] = (byte)((data[0] & 0x00ff) >> 0);     // 書込データ
            //sendData[1] = (byte)((data[0] & 0xff00) >> 8);     // 書込データ
            int ret = WriteMC(start, 1, data, usFrame, usTimer);

            //int ret = WriteMC(start, 1, sendData);
            //var size = Marshal.SizeOf(typeof(Byte)) * num_bit;
            //var ptr = Marshal.AllocCoTaskMem(size);
            //Marshal.Copy(data, 0, ptr, num_bit);
            ////int ret = MClib_memory_area_write_bit(sys, start, ptr, num_bit);
            //int ret = 0;

            ////Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int WriteInt16(String start, ref Int16[] data, int num_int16, ushort usFrame, ushort usTimer)
        {
            byte[] sendData = new byte[2];
            sendData[0] = (byte)((data[0] & 0x00ff) >> 0);     // 書込データ
            sendData[1] = (byte)((data[0] & 0xff00) >> 8);     // 書込データ
            int ret = WriteMC(start, 2, sendData, usFrame, usTimer);

            //int ret = WriteMC(start, 2, sendData);
            //var size = Marshal.SizeOf(typeof(Int16)) * num_int16;
            //var ptr = Marshal.AllocCoTaskMem(size);
            //Marshal.Copy(data, 0, ptr, num_int16);
            ////int ret = MClib_memory_area_write_int16(sys, start, ptr, num_int16);
            //int ret = 0;

            //Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int WriteInt32(String start, ref Int32[] data, int num_int32, ushort usFrame, ushort usTimer)
        {
            byte[] sendData = new byte[4];
            sendData[0] = (byte)((data[0] & 0x000000ff) >> 0);     // 書込データ
            sendData[1] = (byte)((data[0] & 0x0000ff00) >> 8);     // 書込データ
            sendData[2] = (byte)((data[0] & 0x00ff0000) >> 16);     // 書込データ
            sendData[3] = (byte)((data[0] & 0xff000000) >> 24);     // 書込データ
            int ret = WriteMC(start, 4, sendData, usFrame, usTimer);
            //var size = Marshal.SizeOf(typeof(Int32)) * num_int32;
            //var ptr = Marshal.AllocCoTaskMem(size);
            //Marshal.Copy(data, 0, ptr, num_int32);
            ////int ret = MClib_memory_area_write_int32(sys, start, ptr, num_int32);
            //int ret = 0;


            //sendData[19] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x0000ff) >> 0);    // 先頭デバイス
            //sendData[20] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0x00ff00) >> 8);    // 〃
            //sendData[21] = (byte)((int.Parse(m.Groups["Adress"].Value) & 0xff0000) >> 16);   // 〃 (上位)



            //Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int WriteUInt16(String start, ref UInt16[] data, int num_uint16, ushort usFrame, ushort usTimer)
        {
            byte[] sendData = new byte[2];
            sendData[0] = (byte)((data[0] & 0x00ff) >> 0);     // 書込データ
            sendData[1] = (byte)((data[0] & 0xff00) >> 8);     // 書込データ

            int ret = WriteMC(start, 2, sendData, usFrame, usTimer);

            //var size = Marshal.SizeOf(typeof(UInt16)) * num_uint16;
            //var buf = new short[num_uint16];
            //var ptr = Marshal.AllocCoTaskMem(size);
            //var i = 0;
            //foreach (var val in buf)
            //    buf[i] = (short)data[i++];
            //Marshal.Copy(buf, 0, ptr, num_uint16);
            ////int ret = MClib_memory_area_write_uint16(sys, start, ptr, num_uint16);
            //int ret = 0;

            ////Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int WriteUInt32(String start, ref UInt32[] data, int num_uint32, ushort usFrame, ushort usTimer)
        {
            int ret = 0;
            if (num_uint32 == 2)
            {
                byte[] sendData = new byte[8];
                sendData[0] = (byte)((data[0] & 0x000000ff) >> 0);     // 書込データ
                sendData[1] = (byte)((data[0] & 0x0000ff00) >> 8);     // 書込データ
                sendData[2] = (byte)((data[0] & 0x00ff0000) >> 16);     // 書込データ
                sendData[3] = (byte)((data[0] & 0xff000000) >> 24);     // 書込データ
                sendData[4] = (byte)((data[1] & 0x000000ff) >> 0);     // 書込データ
                sendData[5] = (byte)((data[1] & 0x0000ff00) >> 8);     // 書込データ
                sendData[6] = (byte)((data[1] & 0x00ff0000) >> 16);     // 書込データ
                sendData[7] = (byte)((data[1] & 0xff000000) >> 24);     // 書込データ
                ret = WriteMC(start, 8, sendData, usFrame, usTimer);

            }
            else
            {
                byte[] sendData = new byte[4];
                sendData[0] = (byte)((data[0] & 0x000000ff) >> 0);     // 書込データ
                sendData[1] = (byte)((data[0] & 0x0000ff00) >> 8);     // 書込データ
                sendData[2] = (byte)((data[0] & 0x00ff0000) >> 16);     // 書込データ
                sendData[3] = (byte)((data[0] & 0xff000000) >> 24);     // 書込データ
                ret = WriteMC(start, 4, sendData, usFrame, usTimer);

            }
            //var size = Marshal.SizeOf(typeof(Int32)) * num_uint32;
            //var buf = new int[num_uint32];
            //var ptr = Marshal.AllocCoTaskMem(size);
            //var i = 0;
            //foreach (var val in buf)
            //    buf[i] = (int)data[i++];
            //Marshal.Copy(buf ,0 ,ptr, num_uint32);
            ////int ret = MClib_memory_area_write_uint32(sys, start, ptr, num_uint32);
            //int ret = 0;

            //Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int ReadBit(String start, out Byte[] data, int num_bit)
        {
            data = new Byte[num_bit];
            var size = Marshal.SizeOf(typeof(Byte)) * num_bit;
            var ptr = Marshal.AllocCoTaskMem(size);
            //int ret = MClib_memory_area_read_bit(sys, start, ptr, num_bit);
            int ret = 0;

            if (ret == 0)
            {
                Marshal.Copy(ptr, data, 0, num_bit);
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int ReadInt16(String start, out Int16[] data, int num_int16)
        {
            data = new Int16[num_int16];
            var size = Marshal.SizeOf(typeof(Int16)) * num_int16;
            var ptr = Marshal.AllocCoTaskMem(size);
            //var ret = MClib_memory_area_read_int16(sys, start, ptr, num_int16);
            var ret = 0;

            if (ret == 0)
            {
                Marshal.Copy(ptr, data, 0, num_int16);
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int ReadInt32(String start, out Int32[] data, int num_int32)
        {
            data = new Int32[num_int32];
            var size = Marshal.SizeOf(typeof(Int32)) * num_int32;
            var ptr = Marshal.AllocCoTaskMem(size);
            //var ret = MClib_memory_area_read_int32(sys, start, ptr, num_int32);
            var ret = 0;

            if (ret == 0)
            {
                Marshal.Copy(ptr, data, 0, num_int32);
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int ReadUInt16(String start, out UInt16[] data, int num_uint16)
        {
            data = new UInt16[num_uint16];
            var buf = new short[num_uint16];
            var size = Marshal.SizeOf(typeof(UInt16)) * num_uint16;
            var ptr = Marshal.AllocCoTaskMem(size);
            //var ret = MClib_memory_area_read_uint16(sys, start, ptr, num_uint16);
            var ret = 0;

            if (ret == 0)
            {
                Marshal.Copy(ptr, buf, 0, num_uint16);
                var i = 0;
                foreach (var val in buf)
                    data[i++] = (UInt16)val;
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }

        public int ReadUInt32(String start, out UInt32[] data, int num_uint32)
        {
            data = new UInt32[num_uint32];
            var buf = new int[num_uint32];
            var size = Marshal.SizeOf(typeof(UInt32)) * num_uint32;
            var ptr = Marshal.AllocCoTaskMem(size);
            //var ret = MClib_memory_area_read_uint32(sys, start, ptr, num_uint32);
            var ret = 0;
            if (ret == 0)
            {
                Marshal.Copy(ptr, buf, 0, num_uint32);
                var i = 0;
                foreach (var val in data)
                    data[i++] = (UInt32)val;
            }
            Marshal.FreeCoTaskMem(ptr);
            return ret;
        }
        /*----------------------------------------
        ** 【機能】デバイスコードがあるかチェック
        ** 【パラメータ】
        **	 loMemType: デバイスコード
        **   戻り値： 読込要素サイズ＝成功  -1＝非サポートデバイスコード
        **----------------------------------------*/
        byte ChkDevType(string stMemType)
        {
            switch (stMemType)
            {
                case "SM": // 特殊リレー
                    return 0x91;
                case "SD": // 特殊レジスタ(SD)
                    return 0xA9;

                case "X": // 入力(X)
                    return 0x9C;
                case "Y": // 出力(Y)
                    return 0x9D;
                case "M": // 内部リレー(M)
                    return 0x90;
                case "L": // ラッチリレー(L)
                    return 0x92;
                case "F": // アナンシェータ(F)
                    return 0x93;
                case "V": // エッジリレー(V)
                    return 0x94;
                case "B": // リンクリレー(B)
                    return 0xA0;
                case "D": // データレジスタ(D)
                    return 0xA8;
                case "W": // リンクレジスタ(W)
                    return 0xB4;
                case "TS": // 接点(TS)
                    return 0xC1;
                case "TC": // コイル(TC)
                    return 0xC0;
                case "TN": // 現在値(TN)
                    return 0xC2;
                case "STS": // 接点(STS)
                    return 0xC7;
                case "STC": // コイル(STC)
                    return 0xC6;
                case "STN": // 現在値(STN)
                    return 0xC8;
                case "CS": // 接点(CS)
                    return 0xC4;
                case "CC": // コイル(CC)
                    return 0xC3;
                case "CN": // 現在値(CN)
                    return 0xC5;
                case "SB": // リンク特殊リレー(SB)
                    return 0xA1;
                case "SW": // リンク特殊レジスタ(SW)
                    return 0xB5;
                case "S": // ステップリレー(S)
                    return 0x98;
                case "DX": // ダイレクトアクセス入力(DX)
                    return 0xA2;
                case "DY": // ダイレクトアクセス出力(DY)
                    return 0xA3;
                case "Z": // インデックスレジスタ(Z)
                    return 0xCC;
                case "R": // ファイルレジスタ(R)
                    return 0xAF;
                case "ZR": // ファイルレジスタ(ZR)
                    return 0xB0;
                default:
                    return 0;
            }
        }



    }
}
